/*
    TessesFramework a library to make C++ easier for me, used in CrossLang:
   https://git.tesses.org/tesses50/crosslang Copyright (C) 2026 Mike Nolan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#pragma once
#include "../Common.hpp"
namespace Tesses::Framework::Streams {
enum class SeekOrigin : uint8_t { Begin = 0, Current = 1, End = 2 };
class Stream {
  public:
    int32_t ReadByte();
    void WriteByte(uint8_t b);
    virtual bool EndOfStream();
    virtual size_t Read(uint8_t *buff, size_t sz);
    virtual size_t Write(const uint8_t *buff, size_t sz);
    size_t ReadBlock(uint8_t *buff, size_t sz);
    void WriteBlock(const uint8_t *buff, size_t sz);
    virtual bool CanRead();
    virtual bool CanWrite();
    virtual bool CanSeek();
    virtual int64_t GetPosition();
    virtual int64_t GetLength();
    virtual void Flush();
    virtual void Seek(int64_t pos, SeekOrigin whence);
    void CopyTo(std::shared_ptr<Stream> strm, size_t buffSize = 1024);
    void CopyToLimit(std::shared_ptr<Stream> strm, uint64_t len,
                     size_t buffSize = 1024);
    virtual void Close();
    virtual ~Stream();
};
} // namespace Tesses::Framework::Streams